#include "GLApp.h"

namespace
{
    // Global app
    GLApp* g_pApp = NULL;
}

// Pumps message through user defined msg proc from global space
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    if (g_pApp)
    {
        return g_pApp->MsgProc(hwnd, msg, wParam, lParam);
    }
    else
    {
        return DefWindowProc(hwnd, msg, wParam, lParam);
    }
}

GLApp::GLApp()
{
}

GLApp::GLApp(HINSTANCE hInstance)
{
    m_hAppInstance = hInstance;
    m_hAppWnd = NULL;
    m_hDevContext = NULL;
    m_hGLRenderContext = NULL;
    m_ClientWidth = 268;
    m_ClientHeight = 268;
    m_AppTitle = _T("OpenGL Application");
    m_WindowStyle = WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION | WS_MINIMIZEBOX;
    m_FPS = 0.0f;
    g_pApp = this;
}


GLApp::~GLApp()
{
}

UINT GLApp::GetWidth()
{
    return m_ClientWidth;
}

UINT GLApp::GetHeight()
{
    return m_ClientHeight;
}

// Main application loop
int GLApp::Run()
{
    // Calculate timing
    __int64 prevTime = 0;
    QueryPerformanceCounter((LARGE_INTEGER*)&prevTime);

    __int64 countsPerSec = 0;
    QueryPerformanceFrequency((LARGE_INTEGER*)&countsPerSec);
    float secondsPerCount = 1.0f / countsPerSec;

    // Main message loop
    MSG msg = { 0 };
    while (WM_QUIT != msg.message)
    {
        if (PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else
        {
            // Calculate delta time
            __int64 curTime = 0;
            QueryPerformanceCounter((LARGE_INTEGER*)&curTime);
            float deltaTime = (curTime - prevTime) * secondsPerCount;

            // Update
            Update(deltaTime);
            // Render
            Render();
            // Calculate FPS
            CalculateFPS(deltaTime);

            // Swap the buffers
            SwapBuffers(m_hDevContext);

            // Reset prev time for next frame
            prevTime = curTime;
        }
    }

    Shutdown();
    
    return static_cast<int>(msg.wParam);
}

bool GLApp::Init()
{
    if (!InitWindow()) return false;
    if (!InitGL()) return false;

    return true;
}

LRESULT GLApp::MsgProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg)
    {
        case WM_DESTROY:
            PostQuitMessage(0);
            return 0;
        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }
}

bool GLApp::InitWindow()
{
    // WNDCLASSEX
    WNDCLASSEX wcex;
    ZeroMemory(&wcex, sizeof(WNDCLASSEX));
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.cbSize = sizeof(WNDCLASSEX);
    wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.hInstance = m_hAppInstance;
    wcex.lpfnWndProc = MainWndProc;
    wcex.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)GetStockObject(NULL_BRUSH);
    wcex.lpszClassName = _T("GLAPPWNDCLASS");
    wcex.lpszMenuName = NULL; // No menu defined
    wcex.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

    if (!RegisterClassEx(&wcex)) return OutErrorMsg(_T("Failed to register WNDCLASSEX"));

    // Adjust window rect for requested client size
    RECT r = { 0, 0, (LONG)m_ClientWidth, (LONG)m_ClientHeight };
    AdjustWindowRect(&r, m_WindowStyle, FALSE);
    int width = r.right - r.left;
    int height = r.bottom - r.top;
    int x = GetSystemMetrics(SM_CXSCREEN) / 2 - width / 2;
    int y = GetSystemMetrics(SM_CYSCREEN) / 2 - height / 2;

    // Create our window
    m_hAppWnd = CreateWindow(_T("GLAPPWNDCLASS"), m_AppTitle, m_WindowStyle, x, y, width, height, NULL, NULL, m_hAppInstance, NULL);
    if (!m_hAppWnd) return OutErrorMsg(_T("Failed to create window"));

    // Show window
    ShowWindow(m_hAppWnd, SW_SHOW);

    return true;
}

bool GLApp::InitGL()
{
    // Create our device context
    m_hDevContext = GetDC(m_hAppWnd);

    // Create a PIXELFORMATDESCRIPTOR
    PIXELFORMATDESCRIPTOR pfd;
    ZeroMemory(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
    pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 32;
    pfd.cDepthBits = 24;
    pfd.cStencilBits = 8;
    int format = ChoosePixelFormat(m_hDevContext, &pfd);
    if (!SetPixelFormat(m_hDevContext, format, &pfd))
        return OutErrorMsg(_T("Failed to set pixel format"));

    // Create our render context
    m_hGLRenderContext = wglCreateContext(m_hDevContext);
    if (!wglMakeCurrent(m_hDevContext, m_hGLRenderContext))
        return OutErrorMsg(_T("Failed to create and activate render context"));

    // Initialize GLAD
    if (!gladLoadGL())
        return OutErrorMsg(_T("Failed to initialize GLAD"));

    return true;
}

void GLApp::CalculateFPS(float dt)
{
    static float elapsed = 0;
    static int frameCnt = 0;

    elapsed += dt;
    frameCnt++;
    if (elapsed >= 1.0f)
    {
        m_FPS = (float)frameCnt;

        // Output to window title
        std::wstringstream ss;
        ss << "FPS: " << m_FPS;/* << " " <<
            glGetString(GL_VENDOR) << " " << glGetString(GL_RENDERER) << " " << glGetString(GL_VERSION);*/
        TCHAR buffer[256];
        _tcscpy(buffer, ss.str().c_str());
        SetWindowText(m_hAppWnd, buffer);

        elapsed = 0.0f;
        frameCnt = 0;
    }
}

void GLApp::Shutdown()
{
    wglMakeCurrent(NULL, NULL);
    wglDeleteContext(m_hGLRenderContext);
    ReleaseDC(m_hAppWnd, m_hDevContext);
}
